#pragma region License
/*
License (OLC-3)
~~~~~~~~~~~~~~~

Copyright 2024 Joshua Sigona <sigonasr2@gmail.com>

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions or derivations of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

2. Redistributions or derivative works in binary form must reproduce the above
copyright notice. This list of conditions and the following	disclaimer must be
reproduced in the documentation and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its contributors may
be used to endorse or promote products derived from this software without specific
prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS	"AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT	HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL,	SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

Portions of this software are copyright  2024 The FreeType
Project (www.freetype.org). Please see LICENSE_FT.txt for more information.
All rights reserved.
*/
#pragma endregion
#pragma once
#include <iostream>
#include <sstream>
#include <format>
#include <any>
#include <memory>
#include <source_location>

#ifdef _DEBUG
#ifndef __EMSCRIPTEN__
#ifndef __linux__
	#define NEW new ( _NORMAL_BLOCK , __FILE__ , __LINE__ )
	// Replace _NORMAL_BLOCK with _CLIENT_BLOCK if you want the
	// allocations to be of _CLIENT_BLOCK type
#else
	#define NEW new
#endif
#else
	#define NEW new
#endif
#else
    #define NEW new
#endif

#undef ERR //Stupid Windows

#ifndef __EMSCRIPTEN__
	//WARNING! err accepts a stream of data using << operators. If you want to concatenate strings via the + operator, you must wrap the entire operation in ()
	#define ERR(err) { \
		std::stringstream errStream; \
		errStream<<err; \
		Error::log(errStream,std::source_location::current());}

	class Error{
	public:
		inline static void log(std::stringstream&str,std::source_location loc){
			std::cout<<loc.file_name()<<"("<<loc.line()<<":"<<loc.column()<<") "<<loc.function_name()<<": "<<str.str()<<std::endl;
			throw;
		}
	};
#else
	#define ERR(err) { \
		std::stringstream errStream; \
		errStream<<err; \
		Error::log(errStream,std::source_location::current());}

	class Error{
	public:
		inline static void log(std::stringstream&str,std::source_location loc){
			std::cout<<loc.file_name()<<"("<<loc.line()<<":"<<loc.column()<<") "<<loc.function_name()<<": "<<str.str()<<std::endl;
			#ifdef __DEBUG__
				throw;
			#endif
		}
	};

	#define _CrtDumpMemoryLeaks() ((int)0)
#endif

template<typename type>
type DYNAMIC_CAST(auto variable){
	type pointer=dynamic_cast<type>(variable);
	if(pointer==nullptr)ERR("Could not dynamic cast to type "<<typeid(variable).name()<<"!");
	return pointer;
}

template<typename T,typename U>
std::shared_ptr<T>DYNAMIC_POINTER_CAST(const std::shared_ptr<U>&variable){
    std::shared_ptr<T> newVariable=dynamic_pointer_cast<T>(variable);
    if(!newVariable)ERR("Could not dynamic cast to pointer type "<<typeid(newVariable).name()<<"!");
    return newVariable;
}

template<typename T,typename U>
std::shared_ptr<T>DYNAMIC_POINTER_CAST(const std::weak_ptr<U>&variable){
    std::shared_ptr<T> newVariable=dynamic_pointer_cast<T>(variable.lock());
    if(!newVariable)ERR("Could not dynamic cast to pointer type "<<typeid(newVariable).name()<<"!");
    return newVariable;
}