#pragma region License
/*
License (OLC-3)
~~~~~~~~~~~~~~~

Copyright 2024 Joshua Sigona <sigonasr2@gmail.com>

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions or derivations of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

2. Redistributions or derivative works in binary form must reproduce the above
copyright notice. This list of conditions and the following	disclaimer must be
reproduced in the documentation and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its contributors may
be used to endorse or promote products derived from this software without specific
prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS	"AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT	HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL,	SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

Portions of this software are copyright  2024 The FreeType
Project (www.freetype.org). Please see LICENSE_FT.txt for more information.
All rights reserved.
*/
#pragma endregion

#include "TilesetObject.h"
#include "Tileset.h"

void TilesetObject::AddTile(const Tileset&tileset,int tileID){
        if(bounds.size==vi2d{1,1}){
            int tileX=tileID%tileset.columns;
            int tileY=tileID/tileset.columns;
            bounds={vi2d{tileX,tileY}*tileset.tilewidth,{tileset.tilewidth,tileset.tileheight}};
            upperLeftTileId=tileID;
        }else{
            int tileX=tileID%tileset.columns;
            int tileY=tileID/tileset.columns;
            int textureX=tileX*tileset.tilewidth;
            int textureY=tileY*tileset.tileheight;
            int minX=bounds.pos.x;
            int minY=bounds.pos.y;
            int maxX=bounds.pos.x+bounds.size.x;
            int maxY=bounds.pos.y+bounds.size.y;

            if(textureX<minX){
                minX=textureX;
            }
            if(textureX+tileset.tilewidth>maxX){
                maxX=textureX+tileset.tilewidth;
            }
            if(textureY<minY){
                minY=textureY;
            }
            if(textureY+tileset.tileheight>maxY){
                maxY=textureY+tileset.tileheight;
            }
            bounds={{minX,minY},{maxX-minX,maxY-minY}};

            int minTileX=minX/tileset.tilewidth;
            int minTileY=minY/tileset.tileheight;

            int upperLeftTileId=minTileY*tileset.columns+minTileX;
            this->upperLeftTileId=upperLeftTileId;
        }
    }

std::vector<std::pair<int,std::string>>TilesetObject::OutputTag(const Tileset&activeTileset)const{
    std::vector<std::pair<int,std::string>>tileID_tileStr;

    for(int y=bounds.pos.y;y<bounds.bottom().end.y;y+=activeTileset.tileheight){
        for(int x=bounds.pos.x;x<bounds.right().end.x;x+=activeTileset.tilewidth){
            std::string finalStr="";
            int tileID=(y/activeTileset.tileheight)*activeTileset.columns+(x/activeTileset.tilewidth);
            finalStr+=std::format("<tile id=\"{}\" type=\"{}\">",tileID,perspectiveObj?"PerspectiveObject":"Object")+"\n";
            finalStr+="<properties>\n";
            finalStr+=std::format("<property name=\"Name\" value=\"{}\"/>",name)+"\n";
            finalStr+="</properties>\n";
            if(tileID==upperLeftTileId){ //Only the upper left tile will store the collision.
                finalStr+="<objectgroup draworder=\"index\" id=\"1\">\n";
                for(int id=1;const Quadrilateral&quad:collisionTiles){
                    finalStr+=std::format("<object id=\"{}\" x=\"0\" y=\"0\">",id)+"\n";
                    finalStr+=std::format("<polygon points=\"{},{} {},{} {},{} {},{}\"/>",quad[0].x-bounds.pos.x,quad[0].y-bounds.pos.y,quad[1].x-bounds.pos.x,quad[1].y-bounds.pos.y,quad[2].x-bounds.pos.x,quad[2].y-bounds.pos.y,quad[3].x-bounds.pos.x,quad[3].y-bounds.pos.y)+"\n";
                    finalStr+="</object>\n";
                    id++;
                }
                finalStr+="</objectgroup>\n";
            }
            finalStr+="</tile>\n";
            tileID_tileStr.push_back({tileID,finalStr});
        }
    }
    return tileID_tileStr;
}